// General classe for motion based filters

// See legal notice in Copying.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "MVAnalyse.h"
#include "stdio.h"
#include <math.h>

MVAnalyse::MVAnalyse(PClip _child, int size, int pel, int lv, int st, int stp,
                     bool isb, int lambda, int df, bool _mmx, bool _isse, int _idx,
                     IScriptEnvironment* env) :
GenericVideoFilter(_child)
{
	nWidth = vi.width;
	if ( nWidth & 7 ) env->ThrowError("MVAnalyse: Width has to be dividable by 8");

	nHeight = vi.height;
	if ( nHeight & 7 ) env->ThrowError("MVAnalyse: Height has to be dividable by 8");

	nBlkSize = size;
	if (( nBlkSize != 4 ) && ( nBlkSize != 8))
		env->ThrowError("MVAnalyse: Block's size must be 4 or 8");

   nPel = pel;
	if (( nPel != 1 ) && ( nPel != 2 ))
		env->ThrowError("MVAnalyse: pel has to be 1 or 2");

   nDeltaFrame = df;
   if ( nDeltaFrame < 1 )
      nDeltaFrame = 1;
   if ( nDeltaFrame > 8 )
      nDeltaFrame = 8;

	nLvCount = ilog2( ((nWidth / nBlkSize) > (nHeight / nBlkSize)) ? (nHeight / nBlkSize) : (nWidth / nBlkSize) ) - lv;
	nLvCount = ( nLvCount < 1 ) ? 1 : nLvCount;

	isBackward = isb;

   nLambda = lambda;

   //  we'll transmit to the processing filters a handle
   // on the analyzing filter itself ( it's own pointer ), in order
   // to activate the right parameters.
   vi.nchannels = reinterpret_cast<int>(this);

   mmx = _mmx && ( env->GetCPUFlags() & CPUF_MMX );

	isse = _isse && ( env->GetCPUFlags() & CPUF_INTEGER_SSE );

	switch ( st )
	{
	case 0 : 
		searchType = ONETIME;
		nSearchParam = ( stp < 1 ) ? 1 : stp;
		break;
	case 1 :
		searchType = NSTEP;
		nSearchParam = ( stp < 0 ) ? 0 : stp;
		break;
	case 2 : 
		searchType = LOGARITHMIC;
		nSearchParam = ( stp < 1 ) ? 1 : stp;
		break;
	case 3 : 
		searchType = EXHAUSTIVE;
		nSearchParam = ( stp < 0 ) ? 0 : stp;
		break;
	default :
		searchType = LOGARITHMIC;
		nSearchParam = ( stp < 1 ) ? 1 : stp;
		break;
	}

	if (!vi.IsYV12())
		env->ThrowError("MVAnalyse: Needs YV12 data");

   nFlags = 0;
   nFlags |= _isse ? MOTION_USE_ISSE : 0;
   nFlags |= _mmx ? MOTION_USE_MMX : 0;
   nFlags |= isBackward ? MOTION_IS_BACKWARD : 0;

	vectorFields = new GroupOfPlanes(nWidth, nHeight, nBlkSize, nLvCount, nPel, nFlags);

	vi.height = 1;
	vi.width = vectorFields->GetArraySize();
	vi.pixel_type = VideoInfo::CS_BGR32;

   nIdx = _idx;

   nMagicKey = MOTION_MAGIC_KEY;

   MVCore::mvCore.AddFrames(nIdx, 10, nLvCount, nWidth, nHeight, nPel, nBlkSize, nBlkSize, YUVPLANES, _isse);
}

MVAnalyse::~MVAnalyse()
{
	delete vectorFields;
}

PVideoFrame __stdcall MVAnalyse::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame	src = child->GetFrame(n, env);
	const unsigned char *pSrc = src->GetReadPtr(PLANAR_Y);
   const unsigned char *pSrcU = src->GetReadPtr(PLANAR_U);
   const unsigned char *pSrcV = src->GetReadPtr(PLANAR_V);
	int nSrcPitch = src->GetPitch(PLANAR_Y);
   int nSrcPitchUV = src->GetPitch(PLANAR_U);

	PVideoFrame	dst = env->NewVideoFrame(vi);
	unsigned char *pDst = dst->GetWritePtr();

	int minframe = ( isBackward ) ? 0 : nDeltaFrame;
	int maxframe = ( isBackward ) ? vi.num_frames - nDeltaFrame : vi.num_frames;
	int offset = ( isBackward ) ? nDeltaFrame : -nDeltaFrame;

	if (( n < maxframe ) && ( n >= minframe ))
	{
      const unsigned char *pRef = child->GetFrame(n + offset, env)->GetReadPtr(PLANAR_Y);
      const unsigned char *pRefU = child->GetFrame(n + offset, env)->GetReadPtr(PLANAR_U);
      const unsigned char *pRefV = child->GetFrame(n + offset, env)->GetReadPtr(PLANAR_V);
      int nRefPitch = child->GetFrame(n + offset, env)->GetPitch(PLANAR_Y);
      int nRefPitchUV = child->GetFrame(n + offset, env)->GetPitch(PLANAR_U);

      MVFrames *pFrames = MVCore::mvCore.GetFrames(nIdx);

      PROFILE_START(MOTION_PROFILE_INTERPOLATION);
      MVGroupOfFrames *pSrcGOF = pFrames->GetFrame(n);
      pSrcGOF->SetPlane(pSrc, nSrcPitch, YPLANE);
      pSrcGOF->SetPlane(pSrcU, nSrcPitchUV, UPLANE);
      pSrcGOF->SetPlane(pSrcV, nSrcPitchUV, VPLANE);
      pSrcGOF->Reduce();
      pSrcGOF->Pad(YUVPLANES);
      pSrcGOF->Refine(YUVPLANES);

      MVGroupOfFrames *pRefGOF = pFrames->GetFrame(n + offset);
      pRefGOF->SetPlane(pRef, nRefPitch, YPLANE);
      pRefGOF->SetPlane(pRefU, nRefPitchUV, UPLANE);
      pRefGOF->SetPlane(pRefV, nRefPitchUV, VPLANE);
      pRefGOF->Reduce();
      pRefGOF->Pad(YUVPLANES);
      pRefGOF->Refine(YUVPLANES);
      PROFILE_STOP(MOTION_PROFILE_INTERPOLATION);

      vectorFields->SearchMVs(pSrcGOF, pRefGOF, searchType, nSearchParam, nLambda, nFlags, reinterpret_cast<int*>(pDst));

      PROFILE_CUMULATE();
	}
	else vectorFields->WriteDefaultToArray(reinterpret_cast<int*>(pDst));

	return dst;
}