#include "MVBidouille.h"
#include <memory.h>
#include <string.h>
#include <stdio.h>

#include "Interpolation.h"
#include "Padding.h"
#include "CopyCode.h"

MVBidouille::MVBidouille(PClip _child, PClip mvsbw, PClip mvsfw, int _nIdx, const MotionParameters &params, IScriptEnvironment* env) :
GenericMotionFilter(_child, params, mvsbw->GetVideoInfo(), env), mvclipFW(mvsfw), mvclipBW(mvsbw)
{
   CreateFGOP(&fgopBW);
   CreateFGOP(&fgopFW);

   nIdx = _nIdx;

   nOBMCMargin = 0;

   nExtendedWidth = nWidth + 2 * nHPadding + 2 * nOBMCMargin;
   nExtendedHeight = nHeight + 2 * nVPadding + 2 * nOBMCMargin;
   nExtendedPitch = (nExtendedWidth + 15) & (~15);
   nOffset = nHPadding + nOBMCMargin + (nVPadding + nOBMCMargin) * nExtendedPitch;
   nOffsetUV = (nHPadding + nOBMCMargin) / 2 + (nVPadding + nOBMCMargin) * nExtendedPitch / 4;

   pFinalDst = new Pixels*[10];
   pFinalDstU = new Pixels*[10];
   pFinalDstV = new Pixels*[10];

   for ( int i = 0; i < 10; i++ )
   {
      pFinalDst[i] = new Pixels[nExtendedHeight * nExtendedPitch];
      pFinalDstU[i] = new Pixels[nExtendedHeight * nExtendedPitch / 4];
      pFinalDstV[i] = new Pixels[nExtendedHeight * nExtendedPitch / 4];
      pFinalDst[i] += nOffset;
      pFinalDstU[i] += nOffsetUV;
      pFinalDstV[i] += nOffsetUV;
   }

   MVCore::mvCore.AddFrames(nIdx, 10, nLvCount, nWidth, nHeight, nPel, nBlkSize, nBlkSize, YUVPLANES, isse);

   nBlkSadPitch = (nBlkSize + 2 * nOBMCMargin);
   
   pBlkSadModifier = new int *[nPel * nPel];
   for ( int i = 0; i < nPel * nPel; i++ )
      pBlkSadModifier[i] = new int[(nBlkSadPitch + (i&1)) * (nBlkSadPitch + ((i&2)>>1))];

   for ( int k = 0; k < nPel * nPel; k++ )
   {
      int width = nBlkSadPitch + (k&1);
      int height = nBlkSadPitch + ((k&2)>>1);
      for ( int j = 0; j < nOBMCMargin; j++ )
         for ( int i = 0; i < width; i++ )
            pBlkSadModifier[k][i + j * width] = 2;

      for ( int j = nOBMCMargin; j < height - nOBMCMargin; j++ )
      {
         for ( int i = 0; i < nOBMCMargin; i++ )
            pBlkSadModifier[k][i + j * width] = 2;
         for ( int i = nOBMCMargin; i < width - nOBMCMargin; i++ )
            pBlkSadModifier[k][i + j * width] = 1;
         for ( int i = width - nOBMCMargin; i < width; i++ )
            pBlkSadModifier[k][i + j * width] = 2;

         DebugPrintf("%i %i %i %i %i %i %i %i %i %i %i %i\n",
            pBlkSadModifier[k][0 + j * width],
            pBlkSadModifier[k][1 + j * width],
            pBlkSadModifier[k][2 + j * width],
            pBlkSadModifier[k][3 + j * width],
            pBlkSadModifier[k][4 + j * width],
            pBlkSadModifier[k][5 + j * width],
            pBlkSadModifier[k][6 + j * width],
            pBlkSadModifier[k][7 + j * width],
            pBlkSadModifier[k][8 + j * width],
            pBlkSadModifier[k][9 + j * width],
            pBlkSadModifier[k][10 + j * width],
            pBlkSadModifier[k][11 + j * width]);


      }

      for ( int j = height - nOBMCMargin; j < height; j++ )
         for ( int i = 0; i < width; i++ )
            pBlkSadModifier[k][i + j * width] = 2;
   }
}

MVBidouille::~MVBidouille()
{
   for ( int i = 0; i < nPel * nPel; i++ )
   {
      delete[] pBlkSadModifier[i];
   }
   delete[] pBlkSadModifier;

   for ( int i = 0; i < 10; i++ ) 
   {
      pFinalDst[i] -= nOffset;
      pFinalDstU[i] -= nOffsetUV;
      pFinalDstV[i] -= nOffsetUV;
      delete[] pFinalDst[i];
      delete[] pFinalDstU[i];
      delete[] pFinalDstV[i];
   }
   delete[] pFinalDst;
   delete[] pFinalDstU;
   delete[] pFinalDstV;

   
}

void MVBidouille::MoveBlock(MVFrame *pFrame, const FakeBlockData& block, int nRatio, int nIdx)
{
   const unsigned char *pSrc;
   int mvx = (block.GetMV().x * nRatio) >> 10;
   int mvy = (block.GetMV().y * nRatio) >> 10;
   int offx = 0;
   int offy = 0;
   int *pWeights;

   if (nPel == 1)
   {
      pWeights = pBlkSadModifier[0];
      pSrc = pFrame->GetPlane(YPLANE)->GetPointer(block.GetX() - offx - nOBMCMargin, block.GetY() - offy - nOBMCMargin);
   }
   else
   {
      int idx = (mvx & 1) + 2 * (mvy & 1);
      offx = (idx & 1) ? 1 : 0;
      offy = (idx & 2) ? 1 : 0;
      mvx = mvx / 2;
      mvy = mvy / 2;
      pSrc = pFrame->GetPlane(YPLANE)->GetPointer(block.GetX() * 2 - offx - nOBMCMargin * 2, block.GetY() * 2 - offy - nOBMCMargin * 2);
      pWeights = pBlkSadModifier[idx];
   }

   Pixels *pDst = pFinalDst[nIdx] + block.GetX() + mvx - nOBMCMargin + (block.GetY() + mvy - nOBMCMargin) * nExtendedPitch;

   for ( int i = 0; i < nBlkSize + offy + 2 * nOBMCMargin; i++ )
   {
      for ( int j = 0; j < nBlkSize + offx + 2 * nOBMCMargin; j++ )
      {
         pDst[j].nValue = pSrc[j];
         pDst[j].nSad = block.GetSAD() * pWeights[j];
      }
      pDst += nExtendedPitch;
      pSrc += pFrame->GetPlane(YPLANE)->GetPitch();
      pWeights += nBlkSadPitch + offx;
   }
}

void MVBidouille::CollectResults(unsigned char *pDst, int nDstPitch, int width, int height)
{
   unsigned int nMinSad;
   unsigned char bestValue;
   Pixels *pPixels[10];

   for ( int i = 0; i < 10; i++ )
      pPixels[i] = pFinalDst[i];

   for ( int j = 0; j < height; j++ )
   {
      for ( int i = 0; i < width; i++ )
      {
         nMinSad = 0x7fffffff;
         bestValue = 0;
         for ( int k = 0; k < 10; k++ )
         {
            if ( pPixels[k][i].nSad < nMinSad )
            {
               bestValue = pPixels[k][i].nValue;
               nMinSad = pPixels[k][i].nSad;
            }
         }
         pDst[i] = bestValue;
      }
      pDst += nDstPitch;
      for ( int i = 0; i < 10; i++ )
         pPixels[i] += nExtendedPitch;
   }
}

PVideoFrame __stdcall MVBidouille::GetFrame(int n, IScriptEnvironment* env)
{
   PVideoFrame srcprev = child->GetFrame(n, env);
   PVideoFrame srcnext = child->GetFrame(MIN(n + 1, vi.num_frames - 1), env);

   PVideoFrame dst = env->NewVideoFrame(vi);

   GetVectorStream(n, env, mvclipBW, fgopBW);
   GetVectorStream(MIN(n + 1, vi.num_frames - 1), env, mvclipFW, fgopFW);

   DebugPrintf("MVBidouille : vectors validity : %i %i\n",
      fgopBW->GetValidity(), fgopFW->GetValidity());

   pFrames = MVCore::mvCore.GetFrames(nIdx);

   PROFILE_START(MOTION_PROFILE_INTERPOLATION);
   MVGroupOfFrames *pPrevRef = pFrames->GetFrame(n);
   pPrevRef->SetPlane(YRPLAN(srcprev), YPITCH(srcprev), YPLANE);
   pPrevRef->SetPlane(URPLAN(srcprev), UPITCH(srcprev), UPLANE);
   pPrevRef->SetPlane(VRPLAN(srcprev), VPITCH(srcprev), VPLANE);
   pPrevRef->Reduce();
   pPrevRef->Pad(YUVPLANES);
   pPrevRef->Refine(YUVPLANES);

   MVGroupOfFrames *pNextRef = pFrames->GetFrame(MIN(n + 1, vi.num_frames - 1));
   pNextRef->SetPlane(YRPLAN(srcnext), YPITCH(srcnext), YPLANE);
   pNextRef->SetPlane(URPLAN(srcnext), UPITCH(srcnext), UPLANE);
   pNextRef->SetPlane(VRPLAN(srcnext), VPITCH(srcnext), VPLANE);
   pNextRef->Reduce();
   pNextRef->Pad(YUVPLANES);
   pNextRef->Refine(YUVPLANES);
   PROFILE_STOP(MOTION_PROFILE_INTERPOLATION);

   for ( int i = 0; i < 10; i++ )
      memset(pFinalDst[i] - nOffset, 255, nExtendedPitch * nExtendedHeight * sizeof(Pixels));

   for ( int i = 0; i < nBlkCount; i++ )
   {
      MoveBlock(pPrevRef->GetFrame(0), fgopBW->GetPlane(0).GetBlock(i), 512, 0);
      MoveBlock(pNextRef->GetFrame(0), fgopFW->GetPlane(0).GetBlock(i), 512, 1);
   }

   CollectResults(YWPLAN(dst), YPITCH(dst), nWidth, nHeight);

   return dst;
}
