// Make a motion compensate temporal denoiser

// See legal notice in Copying.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "MVCompensate.h"
#include "CopyCode.h"
#include "Padding.h"
#include "Interpolation.h"

MVCompensate::MVCompensate(PClip _child, PClip vectors, bool sc, bool inclip, bool inloop,
                           const MotionParameters &params, IScriptEnvironment* env) :
GenericMotionFilter(_child, params, vectors->GetVideoInfo(), env), mvclip(vectors)
{
	CreateFGOP(&fgop);

   scBehavior = sc;
   inLoop = inloop;
   inClip = inclip;
   nRefPitch = nWidth + nHPadding * 2;
   nRefWidth = nRefPitch;
   nRefHeight = nHeight + nVPadding * 2;

   nRefPitchUV = nRefPitch / 2;
   nRefWidthUV = nRefWidth / 2;
   nRefHeightUV = nRefHeight / 2;

   nOffset = nHPadding + nVPadding * nRefPitch;
   nOffsetUV = nHPadding / 2 + (nVPadding / 2) * nRefPitchUV;

   if ( !inClip && !inLoop )
      SetVectorsNeed(env, vectors, false, false, false, true, true, true);

   if ( inLoop || inClip )
   {
      pRefs = new unsigned char *[nPel * nPel];
      pRefsU = new unsigned char *[nPel * nPel];
      pRefsV = new unsigned char *[nPel * nPel];
      for ( int i = 0; i < nPel * nPel; i++ )
      {
         pRefs[i] = new unsigned char[nRefPitch * nRefHeight];
         pRefsU[i] = new unsigned char[nRefPitchUV * nRefHeightUV];
         pRefsV[i] = new unsigned char[nRefPitchUV * nRefHeightUV];
      }
   }
}

MVCompensate::~MVCompensate()
{
   if ( inLoop || inClip )
   {
      for ( int i = 0; i < nPel * nPel; i++ )
      {
         delete[] pRefs[i];
         delete[] pRefsU[i];
         delete[] pRefsV[i];
      }
      delete[] pRefs;
      delete[] pRefsU;
      delete[] pRefsV;
   }
}

void MVCompensate::MoveBlock(unsigned char *pDst, int nDstPitch, int x, int y, int mvx, int mvy)
{
   pDst += x + y * nDstPitch;
   unsigned char *pSrc;
   if ( nPel == 1 )
      pSrc = pRefs[0] + x + mvx + nHPadding + (y + mvy + nVPadding) * nRefPitch;
   else
   {
	int idx = (mvx & 1) + (mvy & 1) * 2;
      pSrc = pRefs[idx] + (x * 2 + mvx + nHPadding * 2) / 2 + ((y * 2 + mvy + nVPadding * 2) / 2) * nRefPitch;
   }
   for ( int i = 0; i < nBlkSize; i++ )
   {
      memcpy(pDst, pSrc, nBlkSize);
      pDst += nDstPitch;
      pSrc += nRefPitch;
   }
}

void MVCompensate::MoveBlockU(unsigned char *pDst, int nDstPitch, int x, int y, int mvx, int mvy)
{
   pDst += (x / 2) + (y / 2) * nDstPitch;
   unsigned char *pSrc;
   if ( nPel == 1 )
      pSrc = pRefsU[0] + (x + mvx + nHPadding) / 2 + ((y + mvy + nVPadding) / 2) * nRefPitchUV;
   else
   {
	int idx = (mvx & 1) + (mvy & 1) * 2;
      pSrc = pRefsU[idx] + (x * 2 + mvx + nHPadding * 2) / 4 + ((y * 2 + mvy + nVPadding * 2) / 4) * nRefPitchUV;
   }
   for ( int i = 0; i < nBlkSize / 2; i++ )
   {
      memcpy(pDst, pSrc, nBlkSize / 2);
      pDst += nDstPitch;
      pSrc += nRefPitchUV;
   }
}

void MVCompensate::MoveBlockV(unsigned char *pDst, int nDstPitch, int x, int y, int mvx, int mvy)
{
   pDst += (x / 2) + (y / 2) * nDstPitch;
   unsigned char *pSrc;
   if ( nPel == 1 )
      pSrc = pRefsV[0] + (x + mvx + nHPadding) / 2 + ((y + mvy + nVPadding) / 2) * nRefPitchUV;
   else
   {
	int idx = (mvx & 1) + (mvy & 1) * 2;
      pSrc = pRefsV[idx] + (x * 2 + mvx + nHPadding * 2) / 4 + ((y * 2 + mvy + nVPadding * 2) / 4) * nRefPitchUV;
   }
   for ( int i = 0; i < nBlkSize / 2; i++ )
   {
      memcpy(pDst, pSrc, nBlkSize / 2);
      pDst += nDstPitch;
      pSrc += nRefPitchUV;
   }
}

PVideoFrame __stdcall MVCompensate::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame	src	= child->GetFrame(n, env);
   PVideoFrame dst;

   GetVectorStream(n, env, mvclip, fgop);

   int off = ( isBackward ) ? 1 : -1;
   off *= nDeltaFrame;

   if ( IsUsable(fgop) )
   {
      dst = env->NewVideoFrame(vi);
      if ( inClip || inLoop )
      {
         // Copy reference frame into pRefs[0] if we're in inClip mode
         if ( inClip )
         {
            PVideoFrame ref = child->GetFrame(n + off, env);
            env->BitBlt(pRefs[0] + nOffset, nRefPitch, ref->GetReadPtr(PLANAR_Y),
               ref->GetPitch(PLANAR_Y), nWidth, nHeight);
            env->BitBlt(pRefsU[0] + nOffsetUV, nRefPitchUV, ref->GetReadPtr(PLANAR_U),
               ref->GetPitch(PLANAR_U), nWidth / 2, nHeight / 2);
            env->BitBlt(pRefsV[0] + nOffsetUV, nRefPitchUV, ref->GetReadPtr(PLANAR_V),
               ref->GetPitch(PLANAR_V), nWidth / 2, nHeight / 2);
         }

         // Pad the reference frame
         Padding::PadReferenceFrame(pRefs[0], nRefPitch, nHPadding, nVPadding, nWidth, nHeight);
         Padding::PadReferenceFrame(pRefsU[0], nRefPitchUV, nHPadding / 2, nVPadding / 2, nWidth / 2, nHeight / 2);
         Padding::PadReferenceFrame(pRefsV[0], nRefPitchUV, nHPadding / 2, nVPadding / 2, nWidth / 2, nHeight / 2);

         // Resample the reference frame
         if ( nPel == 2 )
         {
            VerticalBilin(pRefs[2], pRefs[0], nRefPitch, nRefPitch, nRefWidth, nRefHeight);
            VerticalBilin(pRefsU[2], pRefsU[0], nRefPitchUV, nRefPitchUV, nRefWidthUV, nRefHeightUV);
            VerticalBilin(pRefsV[2], pRefsV[0], nRefPitchUV, nRefPitchUV, nRefWidthUV, nRefHeightUV);
            HorizontalBilin(pRefs[1], pRefs[0], nRefPitch, nRefPitch, nRefWidth, nRefHeight);
            HorizontalBilin(pRefsU[1], pRefsU[0], nRefPitchUV, nRefPitchUV, nRefWidthUV, nRefHeightUV);
            HorizontalBilin(pRefsV[1], pRefsV[0], nRefPitchUV, nRefPitchUV, nRefWidthUV, nRefHeightUV);
            DiagonalBilin(pRefs[3], pRefs[0], nRefPitch, nRefPitch, nRefWidth, nRefHeight);
            DiagonalBilin(pRefsU[3], pRefsU[0], nRefPitchUV, nRefPitchUV, nRefWidthUV, nRefHeightUV);
            DiagonalBilin(pRefsV[3], pRefsV[0], nRefPitchUV, nRefPitchUV, nRefWidthUV, nRefHeightUV);
         }

         // Makes compensation
         for ( int i = 0; i < nBlkCount; i++ )
         {
            MoveBlock(dst->GetWritePtr(PLANAR_Y), dst->GetPitch(PLANAR_Y), fgop->GetPlane(0)[i].GetX(),
               fgop->GetPlane(0)[i].GetY(), fgop->GetPlane(0)[i].GetMV().x, fgop->GetPlane(0)[i].GetMV().y);
            MoveBlockU(dst->GetWritePtr(PLANAR_U), dst->GetPitch(PLANAR_U), fgop->GetPlane(0)[i].GetX(),
               fgop->GetPlane(0)[i].GetY(), fgop->GetPlane(0)[i].GetMV().x, fgop->GetPlane(0)[i].GetMV().y);
            MoveBlockV(dst->GetWritePtr(PLANAR_V), dst->GetPitch(PLANAR_V), fgop->GetPlane(0)[i].GetX(),
               fgop->GetPlane(0)[i].GetY(), fgop->GetPlane(0)[i].GetMV().x, fgop->GetPlane(0)[i].GetMV().y);
         }

         // Moves the new frame into pRefs[0], if we're in inloop mode
         if ( inLoop )
         {
            env->BitBlt(pRefs[0] + nOffset, nRefWidth, dst->GetReadPtr(PLANAR_Y), dst->GetPitch(PLANAR_Y),
               nWidth, nHeight);
            env->BitBlt(pRefsU[0] + nOffsetUV, nRefWidthUV, dst->GetReadPtr(PLANAR_U), dst->GetPitch(PLANAR_U),
               nWidth / 2, nHeight / 2);
            env->BitBlt(pRefsV[0] + nOffsetUV, nRefWidthUV, dst->GetReadPtr(PLANAR_V), dst->GetPitch(PLANAR_V),
               nWidth / 2, nHeight / 2);
         }
      }
      else
      {
         // we're in the default mode : compensation has been made while searching the mvs
         env->BitBlt(dst->GetWritePtr(PLANAR_Y), dst->GetPitch(PLANAR_Y), 
               fgop->GetCompensatedPlane(), fgop->GetPitch(), 
               nWidth, nHeight);
         env->BitBlt(dst->GetWritePtr(PLANAR_U), dst->GetPitch(PLANAR_U), 
               fgop->GetCompensatedPlaneU(), fgop->GetPitchUV(), 
               nWidth / 2, nHeight / 2);
         env->BitBlt(dst->GetWritePtr(PLANAR_V), dst->GetPitch(PLANAR_V), 
               fgop->GetCompensatedPlaneV(), fgop->GetPitchUV(), 
               nWidth / 2, nHeight / 2);
      }
   }
   else {
      if ( !scBehavior && ( n + off < vi.num_frames ) && ( n + off >= 0 ))
         dst = child->GetFrame(n + off, env);
      else
         dst = src;

      // if we're in in-loop mode, we copy the frame into pRefs[0]
      if ( inLoop )
      {
         env->BitBlt(pRefs[0] + nOffset, nRefWidth, dst->GetReadPtr(PLANAR_Y), dst->GetPitch(PLANAR_Y),
               nWidth, nHeight);
         env->BitBlt(pRefsU[0] + nOffsetUV, nRefWidthUV, dst->GetReadPtr(PLANAR_U), dst->GetPitch(PLANAR_U),
               nWidth / 2, nHeight / 2);
         env->BitBlt(pRefsV[0] + nOffsetUV, nRefWidthUV, dst->GetReadPtr(PLANAR_V), dst->GetPitch(PLANAR_V),
               nWidth / 2, nHeight / 2);
      }
   }
   return dst;
}