
// Make a motion compensate temporal denoiser

// See legal notice in Copying.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "MVDenoise2.h"
#include "CopyCode.h"
#include "Padding.h"

MVDenoise2::MVDenoise2(PClip _child, PClip *mvs, int nbmvs, int _nMode,
                       int thT, int sadT, int thMV, const MotionParameters &params, IScriptEnvironment* env) :
GenericMotionFilter(_child, params, mvs[0]->GetVideoInfo(), env)
{
   nMvCount = nbmvs;
   mvclips = mvs;
   nMode = _nMode;

   fgops = new FakeGroupOfPlanes*[nMvCount];
   for ( int i = 0; i < nMvCount; i++ )
   {
      CreateFGOP(fgops + i);
      SetVectorsNeed(env, mvclips[i], true, true, false, nMode & YPLANE, nMode & UPLANE, nMode & VPLANE);
   }

 	/*
		We give the thresholds in a comprehensive way, so adjustments have to be made
	*/

	TemporalThreshold = thT;
	if ( TemporalThreshold < 0 ) TemporalThreshold = 0;

	SADThreshold = sadT;

	MVThreshold = thMV * thMV * nPel * nPel;

	pitch =  nWidth + 2 * nHPadding;

	offset = nVPadding * pitch + nHPadding;

   size = pitch * ( nHeight + 2 * nVPadding );
}

MVDenoise2::~MVDenoise2()
{
    for ( int i = 0; i < nMvCount; i++ )
        delete fgops[i];
    delete fgops;
}

#define SATURATE(x,a,b) (((x) < (a)) ? (a) : (((x) > (b)) ? (b) : (x)))

void MVDenoise2::DenoiseBlockLuma(unsigned char *pDst, int nDstPitch)
{
   bool *dos = new bool[nMvCount];
   const unsigned char **pRefs = new const unsigned char *[nMvCount];
   int *pitches = new int[nMvCount];
   int *pLumas = new int[nMvCount];

   int nLumaRatio = nBlkSize * nBlkSize;

   for ( int j = 0; j < nMvCount; j++ )
      dos[j] = IsUsable(fgops[j]);
 
	for ( int i = 0; i < nBlkCount; i++ )
	{
      int x = fgops[0]->GetPlane(0)[i].GetX();
		int y = fgops[0]->GetPlane(0)[i].GetY();
      unsigned char *d = pDst + x + y * nDstPitch;
      for ( int j = 0; j < nMvCount; j++ )
      {
         if (dos[j] && (fgops[j]->GetPlane(0)[i].GetSAD() < SADThreshold) && (fgops[j]->GetPlane(0)[i].GetMVLength() < MVThreshold))
         {
            pitches[j] = fgops[j]->GetPitch();
            pRefs[j] = fgops[j]->GetCompensatedPlane() + x + y * pitches[j];
            pLumas[j] = (fgops[j]->GetPlane(0)[i].GetLuma() - fgops[j]->GetPlane(0)[i].GetRefLuma()) / nLumaRatio;
         }
         else
         {
            pitches[j] = nDstPitch;
            pRefs[j] = d;
            pLumas[j] = 0;
         }
      }
      for ( int k = 0; k < nBlkSize; k++ )
		{
			for ( int l = 0; l < nBlkSize; l++ )
			{
				int count = 1;
            int newpix = d[l];
            for ( int j = 0; j < nMvCount; j++ )
            {
               int v = SATURATE(pRefs[j][l] + pLumas[j], 0, 255);
    		      if (MABS(d[l] - v) < TemporalThreshold)
	    	      {
		    	      newpix += v;
				      count += 1;
			      }
            }
	   		d[l] = (newpix + (count >> 1)) / count;
	   	}
	   	d += nDstPitch;
         for ( int j = 0; j < nMvCount; j++ )
            pRefs[j] += pitches[j];
		}
	}
   delete[] dos;
   delete[] pitches;
   delete[] pRefs;
   delete[] pLumas;
}

void MVDenoise2::DenoiseBlockChromaU(unsigned char *pDst, int nDstPitch)
{
   bool *dos = new bool[nMvCount];
   const unsigned char **pRefs = new const unsigned char *[nMvCount];
   int *pitches = new int[nMvCount];

   for ( int j = 0; j < nMvCount; j++ )
      dos[j] = IsUsable(fgops[j]);
 
	for ( int i = 0; i < nBlkCount; i++ )
	{
      int x = fgops[0]->GetPlane(0)[i].GetX() >> 1;
		int y = fgops[0]->GetPlane(0)[i].GetY() >> 1;
      unsigned char *d = pDst + x + y * nDstPitch;
      for ( int j = 0; j < nMvCount; j++ )
      {
         if (dos[j] && (fgops[j]->GetPlane(0)[i].GetSAD() < SADThreshold) && (fgops[j]->GetPlane(0)[i].GetMVLength() < MVThreshold))
         {
            pitches[j] = fgops[j]->GetPitchUV();
            pRefs[j] = fgops[j]->GetCompensatedPlaneU() + x + y * pitches[j];
         }
         else
         {
            pitches[j] = nDstPitch;
            pRefs[j] = d;
         }
      }
      for ( int k = 0; k < nBlkSize >> 1; k++ )
		{
			for ( int l = 0; l < nBlkSize >> 1; l++ )
			{
				int count = 1;
            int newpix = d[l];
            for ( int j = 0; j < nMvCount; j++ )
            {
               int v = SATURATE(pRefs[j][l], 0, 255);
    		      if (MABS(d[l] - v) < TemporalThreshold)
	    	      {
		    	      newpix += v;
				      count += 1;
			      }
            }
	   		d[l] = (newpix + (count >> 1)) / count;
	   	}
	   	d += nDstPitch;
         for ( int j = 0; j < nMvCount; j++ )
            pRefs[j] += pitches[j];
		}
	}
   delete[] dos;
   delete[] pitches;
   delete[] pRefs;
}

void MVDenoise2::DenoiseBlockChromaV(unsigned char *pDst, int nDstPitch)
{
   bool *dos = new bool[nMvCount];
   const unsigned char **pRefs = new const unsigned char *[nMvCount];
   int *pitches = new int[nMvCount];

   for ( int j = 0; j < nMvCount; j++ )
      dos[j] = IsUsable(fgops[j]);
 
	for ( int i = 0; i < nBlkCount; i++ )
	{
      int x = fgops[0]->GetPlane(0)[i].GetX() >> 1;
		int y = fgops[0]->GetPlane(0)[i].GetY() >> 1;
      unsigned char *d = pDst + x + y * nDstPitch;
      for ( int j = 0; j < nMvCount; j++ )
      {
         if (dos[j] && (fgops[j]->GetPlane(0)[i].GetSAD() < SADThreshold) && (fgops[j]->GetPlane(0)[i].GetMVLength() < MVThreshold))
         {
            pitches[j] = fgops[j]->GetPitchUV();
            pRefs[j] = fgops[j]->GetCompensatedPlaneV() + x + y * pitches[j];
         }
         else
         {
            pitches[j] = nDstPitch;
            pRefs[j] = d;
         }
      }
      for ( int k = 0; k < nBlkSize >> 1; k++ )
		{
			for ( int l = 0; l < nBlkSize >> 1; l++ )
			{
				int count = 1;
            int newpix = d[l];
            for ( int j = 0; j < nMvCount; j++ )
            {
               int v = SATURATE(pRefs[j][l], 0, 255);
    		      if (MABS(d[l] - v) < TemporalThreshold)
	    	      {
		    	      newpix += v;
				      count += 1;
			      }
            }
	   		d[l] = (newpix + (count >> 1)) / count;
	   	}
	   	d += nDstPitch;
         for ( int j = 0; j < nMvCount; j++ )
            pRefs[j] += pitches[j];
		}
	}
   delete[] dos;
   delete[] pitches;
   delete[] pRefs;
}

PVideoFrame __stdcall MVDenoise2::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame	src = child->GetFrame(n, env);
	PVideoFrame	dst = env->NewVideoFrame(vi);

	env->BitBlt(UWPLAN(dst), UPITCH(dst), URPLAN(src), UPITCH(src), nWidth >> 1, nHeight >> 1);
   env->BitBlt(VWPLAN(dst), VPITCH(dst), VRPLAN(src), VPITCH(src), nWidth >> 1, nHeight >> 1);
   env->BitBlt(YWPLAN(dst), YPITCH(dst), YRPLAN(src), YPITCH(src), nWidth, nHeight);

   for ( int j = 0; j < nMvCount; j++ )
      GetVectorStream(n, env, mvclips[j], fgops[j]);

   if ( nMode & YPLANE )
      DenoiseBlockLuma(YWPLAN(dst), YPITCH(dst));
   if ( nMode & UPLANE )
      DenoiseBlockChromaU(UWPLAN(dst), UPITCH(dst));
   if ( nMode & VPLANE )
      DenoiseBlockChromaV(VWPLAN(dst), VPITCH(dst));

	return dst;
}