// Show the motion vectors of a clip

// See legal notice in Copying.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "MVShow.h"
#include <stdio.h>

MVShow::MVShow(PClip _child, PClip mvs, int _scale, int _plane, int _tol, bool _showsad,
               const MotionParameters &params, IScriptEnvironment* env) :
GenericMotionFilter(_child, params, mvs->GetVideoInfo(), env), mvclip(mvs)
{
	nScale = _scale;
	if ( nScale < 1 )
      nScale = 1;

	nPlane = _plane;
	if ( nPlane < 0 )
      nPlane = 0;
	if ( nPlane >= nLvCount - 1 )
      nPlane = nLvCount - 1;

	nTolerance = _tol;
   if ( nTolerance < 0 )
      nTolerance = 0;

	showSad = _showsad;

	CreateFGOP(&fgop);
}

MVShow::~MVShow()
{
   delete fgop;
}

PVideoFrame __stdcall MVShow::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame	src = child->GetFrame(n, env);
	PVideoFrame	dst = env->NewVideoFrame(vi);

   GetVectorStream(n, env, mvclip, fgop);

	// Copy the frame into the created frame
   env->BitBlt(UWPLAN(dst), UPITCH(dst), URPLAN(src), UPITCH(src), nWidth >> 1, nHeight >> 1);
   env->BitBlt(VWPLAN(dst), VPITCH(dst), VRPLAN(src), VPITCH(src), nWidth >> 1, nHeight >> 1);
   if ( !nPlane )
	   env->BitBlt(YWPLAN(dst), YPITCH(dst), YRPLAN(src), YPITCH(src), nWidth, nHeight);

   if ( IsUsable(fgop) )
      DrawMVs(YWPLAN(dst), YPITCH(dst), YRPLAN(src), YPITCH(src));

	if ( showSad ) {
		char buf[80];
		double mean = 0;
		for ( int i = 0; i < fgop->GetPlane(nPlane).GetBlockCount(); i++ )
			mean += fgop->GetPlane(nPlane)[i].GetSAD();
		sprintf(buf, "%f", mean / fgop->GetPlane(nPlane).GetBlockCount());
		DrawString(dst, 0, 0, buf);
	}
	return dst;
}

void MVShow::DrawPixel(unsigned char *pDst, int nDstPitch, int x, int y, int w, int h)
{
	if (( x >= 0 ) && ( x < w ) && ( y >= 0 ) && ( y < h ))
		pDst[x + y * nDstPitch] = 255;
}

// Draw the vector, scaled with the right scalar factor.
void MVShow::DrawMV(unsigned char *pDst, int nDstPitch, int scale,
			        int x, int y, int size, int w, int h, VECTOR vector, int pel)
{
	int x0 = (size / 2) + x - 1;
	int y0 = (size / 2) + y - 1;

    bool yLonger = false;
	int incrementVal, endVal;
	int shortLen =  vector.y / pel;
	int longLen = vector.x / pel;
	if ( abs(shortLen) > abs(longLen)) {
		int swap = shortLen;
		shortLen = longLen;
		longLen = swap;
		yLonger = true;
	}
	endVal = longLen;
	if (longLen < 0) {
		incrementVal = -1;
		longLen = -longLen;
	} else incrementVal = 1;
	int decInc;
	if (longLen==0) decInc=0;
	else decInc = (shortLen << 16) / longLen;
	int j=0;
	if (yLonger) {	
		for (int i = 0; i != endVal; i += incrementVal) {
            DrawPixel(pDst, nDstPitch, x0 + (j >> 16), y0 + i, w, h);
			j += decInc;
		}
	} else {
		for (int i = 0; i != endVal; i += incrementVal) {
            DrawPixel(pDst, nDstPitch, x0 + i, y0 + (j>> 16), w, h);
			j += decInc;
		}
	}
}

void MVShow::DrawMVs(unsigned char *pDst, int nDstPitch, const unsigned char *pSrc,
                     int nSrcPitch)
{
	const FakePlaneOfBlocks &plane = fgop->GetPlane(nPlane);
	int effectiveScale = plane.GetEffectiveScale();

	const unsigned char *s = pSrc;
	
    if ( nPlane )
    {
	    for ( int j = 0; j < plane.GetHeight(); j++)
	    {
		    for ( int i = 0; i < plane.GetWidth(); i++)
		    {
			    int val = 0;
			    for ( int x = i * effectiveScale ; x < effectiveScale * (i + 1); x++ )
				    for ( int y = j * effectiveScale; y < effectiveScale * (j + 1); y++ )
					    val += pSrc[x + y * nSrcPitch];
			    val = (val + effectiveScale * effectiveScale / 2) / (effectiveScale * effectiveScale);
			    for ( int x = i * effectiveScale ; x < effectiveScale * (i + 1); x++ )
				    for ( int y = j * effectiveScale; y < effectiveScale * (j + 1); y++ )
					    pDst[x + y * nDstPitch] = val;
		    }
		    s += effectiveScale * nSrcPitch;
	    }
    }
	for ( int i = 0; i < plane.GetBlockCount(); i++ )
		if ( plane[i].GetSAD() < nTolerance )
			DrawMV(pDst, nDstPitch, nScale * effectiveScale, plane[i].GetX() * effectiveScale, 
				plane[i].GetY() * effectiveScale, plane.GetBlockSize() * effectiveScale,
                plane.GetWidth() * effectiveScale, plane.GetHeight() * effectiveScale, 
                plane[i].GetMV(), plane.GetPel());
}