// See legal notice in Copying.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "PlaneOfBlocks.h"
#include "Padding.h"


PlaneOfBlocks::PlaneOfBlocks(int _nBlkX, int _nBlkY, int _nBlkSize, int _nPel, int _nLevel, int _nFlags)
{

/* constant fields */

   nPel = _nPel;
   nLogPel = ilog2(nPel);

   nLogScale = _nLevel;
   nScale = iexp2(nLogScale);

   nBlkSize = _nBlkSize;

   nBlkX = _nBlkX;
   nBlkY = _nBlkY;
   nBlkCount = nBlkX * nBlkY;

   nFlags = _nFlags;

   smallestPlane = (bool)(nFlags & MOTION_SMALLEST_PLANE);
   mmx = (bool)(nFlags & MOTION_USE_MMX);
   isse = (bool)(nFlags & MOTION_USE_ISSE);

/* arrays memory allocation */

   vectors = new VECTOR[nBlkCount];
 
/* function's pointers initialization */

   if ( isse )
   {
      if ( nBlkSize == 4 )
      {
         SAD = SAD4x4_iSSE;
         VAR = Var4x4_iSSE;
         LUMA = Luma4x4_iSSE;
         BLITLUMA = Copy4_mmx;
         BLITCHROMA = Copy2_mmx;
         SADCHROMA = SAD2x2_C;
      }
      else
      {
         SAD = SAD8x8_iSSE;
         VAR = Var8x8_iSSE;
         LUMA = Luma8x8_iSSE;
         BLITLUMA = Copy8_mmx;
         BLITCHROMA = Copy4_mmx;
         SADCHROMA = SAD4x4_iSSE;
      }
   }
   else
   {
      if ( nBlkSize == 4 )
      {
         SAD = SAD4x4_C;
         VAR = Var4x4_C;
         LUMA = Luma4x4_C;
         BLITLUMA = Copy4_c; // "mmx" version could be used, but it's more like a debugging version
         BLITCHROMA = Copy2_c; // idem
         SADCHROMA = SAD2x2_C;
      }
      else
      {
         SAD = SAD8x8_C;
         VAR = Var8x8_C;
         LUMA = Luma8x8_C;
         BLITLUMA = Copy8_c;
         BLITCHROMA = Copy4_c; // idem
         SADCHROMA = SAD4x4_C;
      }
   }
}

PlaneOfBlocks::~PlaneOfBlocks()
{
   delete[] vectors;
}

void PlaneOfBlocks::SearchMVs(MVFrame *_pSrcFrame, MVFrame *_pRefFrame,
                              SearchType st, int stp, int lambda, int flags, int *out)
{
   bool calcSrcLuma = (bool)(flags & MOTION_CALC_SRC_LUMA);
   bool calcRefLuma = (bool)(flags & MOTION_CALC_REF_LUMA);
   bool calcSrcVar = (bool)(flags & MOTION_CALC_VAR);
   int nOutPitchY = nBlkX * nBlkSize;
   int nOutPitchUV = nBlkX * nBlkSize / 2;

   // write the plane's header
   WriteHeaderToArray(out);

   int *pBlkData = out + 1;

   Uint8 *pMoCompDataY, *pMoCompDataU, *pMoCompDataV;
   if ( nLogScale == 0 )
   {
      pMoCompDataY = reinterpret_cast<Uint8 *>(out + out[0] + 1);
      pMoCompDataU = pMoCompDataY + nBlkX * nBlkY * nBlkSize * nBlkSize;
      pMoCompDataV = pMoCompDataU + nBlkX * nBlkY * nBlkSize * nBlkSize / 4;
      out += out[0];
      out[0] = nBlkX * nBlkY * nBlkSize * nBlkSize * 3 / 8 + 1;
   }

   nFlags |= flags;

   pSrcFrame = _pSrcFrame;
   pRefFrame = _pRefFrame;

   x[0] = pSrcFrame->GetPlane(YPLANE)->GetHPadding();
   x[1] = pSrcFrame->GetPlane(UPLANE)->GetHPadding();
   x[2] = pSrcFrame->GetPlane(VPLANE)->GetHPadding();
	y[0] = pSrcFrame->GetPlane(YPLANE)->GetVPadding();
   y[1] = pSrcFrame->GetPlane(UPLANE)->GetVPadding();
   y[2] = pSrcFrame->GetPlane(VPLANE)->GetVPadding();

	blkx = 0;
	blky = 0;

   nSrcPitch[0] = pSrcFrame->GetPlane(YPLANE)->GetPitch();
   nSrcPitch[1] = pSrcFrame->GetPlane(UPLANE)->GetPitch();
   nSrcPitch[2] = pSrcFrame->GetPlane(VPLANE)->GetPitch();
   nRefPitch[0] = pRefFrame->GetPlane(YPLANE)->GetPitch();
   nRefPitch[1] = pRefFrame->GetPlane(UPLANE)->GetPitch();
   nRefPitch[2] = pRefFrame->GetPlane(VPLANE)->GetPitch();

   searchType = st;
	nSearchParam = stp;
   lambda = lambda / (nPel * nPel);

	// Functions using float must not be used here
	for ( blkIdx = 0; blkIdx < nBlkCount; blkIdx++ )
	{
      PROFILE_START(MOTION_PROFILE_ME);

      pSrc[0] = pSrcFrame->GetPlane(YPLANE)->GetAbsolutePelPointer(x[0], y[0]);
      pSrc[1] = pSrcFrame->GetPlane(UPLANE)->GetAbsolutePelPointer(x[1], y[1]);
      pSrc[2] = pSrcFrame->GetPlane(VPLANE)->GetAbsolutePelPointer(x[2], y[2]);

      if ( blky == 0 ) nLambda = 0;
      else nLambda = lambda;

      /* computes search boundaries */
      nDxMax = nPel * (pSrcFrame->GetPlane(YPLANE)->GetExtendedWidth() - x[0] - nBlkSize);
      nDyMax = nPel * (pSrcFrame->GetPlane(YPLANE)->GetExtendedHeight()  - y[0] - nBlkSize);
	   nDxMin = -nPel * x[0];
		nDyMin = -nPel * y[0];

      /* search the mv */
		predictor = ClipMV(vectors[blkIdx]);
 
      PseudoEPZSearch();
            
      /* write the results */
      pBlkData[0] = bestMV.x;
      pBlkData[1] = bestMV.y;
      pBlkData[2] = bestMV.sad;
      PROFILE_STOP(MOTION_PROFILE_ME);

      PROFILE_START(MOTION_PROFILE_LUMA_VAR);
      if ( calcSrcVar )
         pBlkData[3] = VAR(pSrc[0], nSrcPitch[0], pBlkData + 4);
      else if ( calcSrcLuma )
      {
         pBlkData[3] = 0;
         pBlkData[4] = LUMA(pSrc[0], nSrcPitch[0]);
      }
      else 
         pBlkData[3] = pBlkData[4] = 0;

      if ( calcRefLuma )
         pBlkData[5] = LUMA(GetRefBlock(bestMV.x, bestMV.y), nRefPitch[0]);
      else
         pBlkData[5] = 0;
      pBlkData += 6;
      PROFILE_STOP(MOTION_PROFILE_LUMA_VAR);

      PROFILE_START(MOTION_PROFILE_COMPENSATION);
      if ( nLogScale == 0 )
      {
         if ( nFlags & MOTION_COMPENSATE_LUMA )
            BLITLUMA(pMoCompDataY, nOutPitchY, GetRefBlock(bestMV.x, bestMV.y), nRefPitch[0]);
         if ( nFlags & MOTION_COMPENSATE_CHROMA_U )
            BLITCHROMA(pMoCompDataU, nOutPitchUV, GetRefBlockU(bestMV.x, bestMV.y), nRefPitch[1]);
         if ( nFlags & MOTION_COMPENSATE_CHROMA_V )
            BLITCHROMA(pMoCompDataV, nOutPitchUV, GetRefBlockV(bestMV.x, bestMV.y), nRefPitch[2]);
      }
      PROFILE_STOP(MOTION_PROFILE_COMPENSATION);
  
        /* increment indexes & pointers */
		blkx++;
      x[0] += nBlkSize;
      x[1] += nBlkSize >> 1;
      x[2] += nBlkSize >> 1;
      pMoCompDataY += nBlkSize;
      pMoCompDataU += nBlkSize >> 1;
      pMoCompDataV += nBlkSize >> 1;
 		if ( blkx == nBlkX )
		{
			blkx = 0;
         x[0] = pSrcFrame->GetPlane(YPLANE)->GetHPadding();
         x[1] = pSrcFrame->GetPlane(UPLANE)->GetHPadding();
         x[2] = pSrcFrame->GetPlane(VPLANE)->GetHPadding();
			blky++;
			y[0] += nBlkSize;
         y[1] += nBlkSize >> 1;
         y[2] += nBlkSize >> 1;
         pMoCompDataY += (nBlkSize - 1) * nOutPitchY;
         pMoCompDataU += (nBlkSize / 2 - 1) * nOutPitchUV;
         pMoCompDataV += (nBlkSize / 2 - 1) * nOutPitchUV;
		}
	}

	__asm { emms }
}

void PlaneOfBlocks::InterpolatePrediction(const PlaneOfBlocks &pob)
{
   int normFactor = 3 - nLogPel + pob.nLogPel;
	int mulFactor = (normFactor < 0) ? -normFactor : 0;
	normFactor = (normFactor < 0) ? 0 : normFactor;

   for ( int l = 0, index = 0; l < nBlkY; l++ )
	{
		for ( int k = 0; k < nBlkX; k++, index++ )
		{
			VECTOR v1, v2, v3, v4;
			int i = k; 
			int j = l;
			if ( i == 2 * pob.nBlkX ) i--;
			if ( j == 2 * pob.nBlkY ) j--;
			int offy = -1 + 2 * ( j % 2);
			int offx = -1 + 2 * ( i % 2);
	
			if (( i == 0 ) || (i == 2 * pob.nBlkX - 1))
			{
				if (( j == 0 ) || ( j == 2 * pob.nBlkY - 1))
				{
					v1 = v2 = v3 = v4 = pob.vectors[i / 2 + (j / 2) * pob.nBlkX];
				}
				else 
				{
					v1 = v2 = pob.vectors[i / 2 + (j / 2) * pob.nBlkX];
					v3 = v4 = pob.vectors[i / 2 + (j / 2 + offy) * pob.nBlkX];
				}
			}
			else if (( j == 0 ) || ( j >= 2 * pob.nBlkY - 1))
			{
				v1 = v2 = pob.vectors[i / 2 + (j / 2) * pob.nBlkX];
				v3 = v4 = pob.vectors[i / 2 + offx + (j / 2) * pob.nBlkX];
			}
			else
			{
				v1 = pob.vectors[i / 2 + (j / 2) * pob.nBlkX];
				v2 = pob.vectors[i / 2 + offx + (j / 2) * pob.nBlkX];
				v3 = pob.vectors[i / 2 + (j / 2 + offy) * pob.nBlkX];
				v4 = pob.vectors[i / 2 + offx + (j / 2 + offy) * pob.nBlkX];
			}

			vectors[index].x = 9 * v1.x + 3 * v2.x + 3 * v3.x + v4.x;
			vectors[index].y = 9 * v1.y + 3 * v2.y + 3 * v3.y + v4.y;
         vectors[index].sad = 9 * v1.sad + 3 * v2.sad + 3 * v3.sad + v4.sad + 8;

			vectors[index].x = (vectors[index].x >> normFactor) << mulFactor;
			vectors[index].y = (vectors[index].y >> normFactor) << mulFactor;
         vectors[index].sad = vectors[index].sad >> 4;
		}
	}
}

void PlaneOfBlocks::WriteHeaderToArray(int *array)
{
    array[0] = nBlkCount * 6 + 1;
}

int PlaneOfBlocks::WriteDefaultToArray(int *array)
{
   array[0] = nBlkCount * 6 + 1;
	memset(array + 1, 0, nBlkCount * sizeof(int) * 6);
   if ( nLogScale == 0 )
   {
      array += array[0];
      array[0] = nBlkX * nBlkY * nBlkSize * nBlkSize * 3 / 8 + 1;
      memset(array + 1, 0, nBlkX * nBlkY * nBlkSize * nBlkSize * 3 / 8);
   }
   return GetArraySize();
}

int PlaneOfBlocks::GetArraySize()
{
	int size = 0;
	size += 1;              // mb data size storage
	size += nBlkCount * 6;  // vectors, sad, luma src, luma ref, var
   
   if ( nLogScale == 0 )
   {
      size += 1;           // compensation size storage
      size += nBlkX * nBlkY * nBlkSize * nBlkSize / 4; // luma comp
      size += nBlkX * nBlkY * nBlkSize * nBlkSize / 16; // chroma u comp
      size += nBlkX * nBlkY * nBlkSize * nBlkSize / 16; // chroma v comp
   }

   return size;
}

void PlaneOfBlocks::FetchPredictors()
{
	// Left predictor
	if ( blkx > 0 ) predictors[1] = ClipMV(vectors[blkIdx - 1]);
	else predictors[1] = zeroMV;

	// Up predictor
	if ( blky > 0 ) predictors[2] = ClipMV(vectors[blkIdx - nBlkX]);
	else predictors[2] = zeroMV;

	// Up-right predictor
	if (( blky > 0 ) && ( blkx < nBlkX - 1 ))
		predictors[3] = ClipMV(vectors[blkIdx - nBlkX + 1]);
	else predictors[3] = zeroMV;

	// Median predictor
	if ( blky > 1 )
	{
		predictors[0].x = Median(predictors[1].x, predictors[2].x, predictors[3].x);
		predictors[0].y = Median(predictors[1].y, predictors[2].y, predictors[3].y);
      predictors[0].sad = Median(predictors[1].sad, predictors[2].sad, predictors[3].sad);
	}
	else {
		predictors[0].x = predictors[1].x + predictors[2].x + predictors[3].x;
		predictors[0].y = predictors[1].y + predictors[2].y + predictors[3].y;
      predictors[0].sad = predictors[1].sad + predictors[2].sad + predictors[3].sad;
	}

    // if there are no other planes, predictor is the median
	if ( smallestPlane ) predictor = predictors[0];
   else 
   {
      if ( predictors[0].sad < predictor.sad )
      {
         predictors[4] = predictor;
         predictor = predictors[0];
         predictors[0] = predictors[4];
      }
   }
   if ( predictor.sad > 400 ) nLambda = 0;
}

void PlaneOfBlocks::PseudoEPZSearch()
{
	int sad;

	FetchPredictors();

	// We treat zero alone
	// Do we bias zero with not taking into account distorsion ?
   bestMV = zeroMV;
	bestMV.sad = SAD(pSrc[0], GetRefBlock(0, 0), nSrcPitch[0], nRefPitch[0])
      + SADCHROMA(pSrc[1], GetRefBlockU(0, 0), nSrcPitch[1], nRefPitch[1])
      + SADCHROMA(pSrc[2], GetRefBlockV(0, 0), nSrcPitch[2], nRefPitch[2]);
	nMinCost = bestMV.sad;

	// Then, the predictor :
	if ( (( predictor.x != 0 ) || ( predictor.y != 0 )) &&
		IsVectorOK(predictor.x, predictor.y) )
	{
		sad = SAD(pSrc[0], GetRefBlock(predictor.x, predictor.y), nSrcPitch[0], nRefPitch[0])
         + SADCHROMA(pSrc[1], GetRefBlockU(predictor.x, predictor.y), nSrcPitch[1], nRefPitch[1])
         + SADCHROMA(pSrc[2], GetRefBlockV(predictor.x, predictor.y), nSrcPitch[2], nRefPitch[2]);
		if ( sad  < nMinCost )
		{
			bestMV.x = predictor.x;
			bestMV.y = predictor.y;
         bestMV.sad = sad;
			nMinCost = sad;
		}
	}

   // then all the other predictors
	for ( int i = 0; i < 4; i++ )
		CheckMV(predictors[i].x, predictors[i].y);

   // then, we refine, according to the search type
	if ( searchType & EXHAUSTIVE )
       ExhaustiveSearch(nSearchParam);

	if ( searchType & LOGARITHMIC )
		for ( int i = nSearchParam; i > 0; i /= 2 )
			DiamondSearch(i);

	if ( searchType & ONETIME )
		for ( int i = nSearchParam; i > 0; i /= 2 )
			OneTimeSearch(i);

	if ( searchType & NSTEP )
		NStepSearch(nSearchParam);
	
	if ( searchType & SQUARE )
		SquareSearch();

   // we store the result
	vectors[blkIdx] = bestMV;
}

void PlaneOfBlocks::DiamondSearch(int length)
{
	// The meaning of the directions are the following :
	//		* 1 means right
	//		* 2 means left
	//		* 4 means down
	//		* 8 means up
	// So 1 + 4 means down right, and so on...

	int dx;
	int dy;

	// We begin by making no assumption on which direction to search.
	int direction = 15;

	int lastDirection;

	while ( direction > 0 )
	{
		dx = bestMV.x;
		dy = bestMV.y;
		lastDirection = direction;
		direction = 0;

		// First, we look the directions that were hinted by the previous step
		// of the algorithm. If we find one, we add it to the set of directions
		// we'll test next
		if ( lastDirection & 1 ) CheckMV2(dx + length, dy, &direction, 1);
		if ( lastDirection & 2 ) CheckMV2(dx - length, dy, &direction, 2);
		if ( lastDirection & 4 ) CheckMV2(dx, dy + length, &direction, 4);
		if ( lastDirection & 8 ) CheckMV2(dx, dy - length, &direction, 8);

		// If one of the directions improves the SAD, we make further tests 
		// on the diagonals
		if ( direction ) {
			lastDirection = direction;
			dx = bestMV.x;
			dy = bestMV.y;

			if ( lastDirection & 3 ) 
			{
				CheckMV2(dx, dy + length, &direction, 4);
				CheckMV2(dx, dy - length, &direction, 8);
			}
			else {
				CheckMV2(dx + length, dy, &direction, 1);
				CheckMV2(dx - length, dy, &direction, 2);
			}
		}

		// If not, we do not stop here. We infer from the last direction the
		// diagonals to be checked, because we might be lucky.
		else {
			switch ( lastDirection ) {
				case 1 : 
					CheckMV2(dx + length, dy + length, &direction, 1 + 4);
					CheckMV2(dx + length, dy - length, &direction, 1 + 8);
					break;
				case 2 : 
					CheckMV2(dx - length, dy + length, &direction, 2 + 4);
					CheckMV2(dx - length, dy - length, &direction, 2 + 8);
					break;
				case 4 : 
					CheckMV2(dx + length, dy + length, &direction, 1 + 4);
					CheckMV2(dx - length, dy + length, &direction, 2 + 4);
					break;
				case 8 : 
					CheckMV2(dx + length, dy - length, &direction, 1 + 8);
					CheckMV2(dx - length, dy - length, &direction, 2 + 8);
					break;
				case 1 + 4 :
					CheckMV2(dx + length, dy + length, &direction, 1 + 4);
					CheckMV2(dx - length, dy + length, &direction, 2 + 4);
					CheckMV2(dx + length, dy - length, &direction, 1 + 8);
					break;
				case 2 + 4 :
					CheckMV2(dx + length, dy + length, &direction, 1 + 4);
					CheckMV2(dx - length, dy + length, &direction, 2 + 4);
					CheckMV2(dx - length, dy - length, &direction, 2 + 8);
					break;
				case 1 + 8 :
					CheckMV2(dx + length, dy + length, &direction, 1 + 4);
					CheckMV2(dx - length, dy - length, &direction, 2 + 8);
					CheckMV2(dx + length, dy - length, &direction, 1 + 8);
					break;
				case 2 + 8 :
					CheckMV2(dx - length, dy - length, &direction, 2 + 8);
					CheckMV2(dx - length, dy + length, &direction, 2 + 4);
					CheckMV2(dx + length, dy - length, &direction, 1 + 8);
					break;
				default :
					// Even the default case may happen, in the first step of the 
					// algorithm for example.
					CheckMV2(dx + length, dy + length, &direction, 1 + 4);
					CheckMV2(dx - length, dy + length, &direction, 2 + 4);
					CheckMV2(dx + length, dy - length, &direction, 1 + 8);
					CheckMV2(dx - length, dy - length, &direction, 2 + 8);
					break;
			}
		}
	}
}
	
void PlaneOfBlocks::SquareSearch()
{
	ExhaustiveSearch(1);
}

void PlaneOfBlocks::ExhaustiveSearch(int s)
{
	int i, j;
	VECTOR mv = bestMV;

	for ( i = -s + 1; i < 0; i++ )
		for ( j = -s + 1; j < s; j++ )
			CheckMV(mv.x + i, mv.y + j);

	for ( i = 1; i < s; i++ )
		for ( j = -s + 1; j < s; j++ )
			CheckMV(mv.x + i, mv.y + j);

	for ( j = -s + 1; j < 0; j++ )
		CheckMV(mv.x, mv.y + j);

	for ( j = 1; j < s; j++ )
		CheckMV(mv.x, mv.y + j);

}

void PlaneOfBlocks::NStepSearch(int stp)
{
	int dx, dy;
	int length = stp;
	while ( length > 0 )
	{
		dx = bestMV.x;
		dy = bestMV.y;

		CheckMV(dx + length, dx + length);
		CheckMV(dx + length, dx);
		CheckMV(dx + length, dx - length);
		CheckMV(dx, dx - length);
		CheckMV(dx, dx + length);
		CheckMV(dx - length, dx + length);
		CheckMV(dx - length, dx);
		CheckMV(dx - length, dx - length);

		length--;
	}
}

void PlaneOfBlocks::OneTimeSearch(int length)
{
	int direction = 0;
	int dx = bestMV.x;
	int dy = bestMV.y;
	
	CheckMV2(dx - length, dy, &direction, 2);
	CheckMV2(dx + length, dy, &direction, 1);

	if ( direction == 1 )
	{
		while ( direction )
		{
			direction = 0;
			dx += length;
			CheckMV2(dx + length, dy, &direction, 1);
		}
	}
	else if ( direction == 2 )
	{
		while ( direction )
		{
			direction = 0;
			dx -= length;
			CheckMV2(dx - length, dy, &direction, 1);
		}
	}

	CheckMV2(dx, dy - length, &direction, 2);
	CheckMV2(dx, dy + length, &direction, 1);

	if ( direction == 1 )
	{
		while ( direction )
		{
			direction = 0;
			dy += length;
			CheckMV2(dx, dy + length, &direction, 1);
		}
	}
	else if ( direction == 2 )
	{
		while ( direction )
		{
			direction = 0;
			dy -= length;
			CheckMV2(dx, dy - length, &direction, 1);
		}
	}
}