// Functions that computes distances between blocks

// See legal notice in Copying.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

/*! \file SADFunctions.h
 *  \brief Calculate the sum of absolute differences between two blocks.
 * 
 *	The sum of absolute differences between two blocks represents the distance
 *	between these two blocks. The lower this value, the closer the blocks will look like
 *  There are two versions, one in plain C, one in iSSE assembler.
 */

#ifndef __SAD_FUNC__
#define __SAD_FUNC__

#include "portability.h"
#include <math.h>
#include "Resizer.h"

/*! \brief Calculate the sum of absolute differences between two blocks.
 *	
 *	\param s First block's pointer.
 *	\param d Second block's pointer.
 *  \param pitch First block's pitch.
 *	\param pitchd Second block's pitch.
 */
typedef unsigned int (SADFunction)(const unsigned char *s, const unsigned char *d,
								           int pitch, int pitchd);

/*! \brief Return the absolute value of x.
 *	\param x 
 */
MOTION_INLINE static unsigned int SADABS(const int x)
{
	return ( x < 0 ) ? -x : x;
}

/*! \brief Experimental function : we change the SAD for SSIM when calculating distance.
 *	So far, results with this change are not clearly better than with the classical SAD,
 *	and the speed loss is great, so it is here only for test purpose.
 */
//SADFunction SSIM8_C;

static unsigned int SAD8x8_C(const unsigned char *s, const unsigned char *d,
					    int pitch,  int pitchd)
{
	unsigned int sum = 0;
	int modulo = pitch - 8;
	int modulod = pitchd - 8;
	for ( int x = 0; x < 8; x++ )
	{
		for ( int y = 0; y < 8; y++ )
		{
			sum += SADABS((*s) - (*d));
			s++; d++;
		}
		s += modulo;
		d += modulod;
	}
	return sum;
}

static unsigned int SAD4x4_C(const unsigned char *s, const unsigned char *d,
					    int pitch,   int pitchd)
{
	unsigned int sum = 0;
	int modulo = pitch - 4;
	int modulod = pitchd - 4;
	for ( int x = 0; x < 4; x++ )
	{
		for ( int y = 0; y < 4; y++ )
		{
			sum += SADABS((*s) - (*d));
			s++; d++;
		}
		s += modulo;
		d += modulod;
	}
	return sum;
}

static unsigned int SAD2x2_C(const unsigned char *pSrc, const unsigned char *pRef,
                             int nSrcPitch, int nRefPitch)
{
 	unsigned int nSum = 0;
   nSum += abs(pSrc[0] - pRef[0]);
   nSum += abs(pSrc[1] - pRef[1]);
   nSum += abs(pSrc[nSrcPitch] - pRef[nRefPitch]);
   nSum += abs(pSrc[nSrcPitch + 1] - pRef[nRefPitch + 1]);
   return nSum;
}

extern "C" unsigned int __cdecl SAD8x8_iSSE_3DNow(const unsigned char * cur,
										  const unsigned char * ref,
										    int cpitch,
										    int rpitch);

extern "C" unsigned int __cdecl SAD8x8_iSSE(const unsigned char * cur,
									const unsigned char * ref,
									  int cpitch,
									  int rpitch);

extern "C" unsigned int __cdecl SAD8x4_iSSE(const unsigned char * cur,
									const unsigned char * ref,
									  int cpitch,
									  int rpitch);

extern "C" unsigned int __cdecl SAD4x8_iSSE(const unsigned char * cur,
									const unsigned char * ref,
									  int cpitch,
									  int rpitch);

extern "C" unsigned int __cdecl SAD4x4_iSSE(const unsigned char * cur,
									const unsigned char * ref,
									  int cpitch,
									  int rpitch);


#endif